from http import HTTPStatus
from typing import NamedTuple, Optional, Union

from .constances import default_headers
from .structures import CaseInsensitiveDict
from .types import ErrorType


class ErrorResponse(NamedTuple):
    """Response for error.

    Used by response_types.
    Used by grant_types.
    """

    error: ErrorType
    description: str
    error_uri: str = ""


class AuthorizationCodeResponse(NamedTuple):
    """Response for authorization_code.

    Used by response_types:
        - ResponseTypeAuthorizationCode
    """

    code: str
    scope: str


class TokenResponse(NamedTuple):
    """Response for token.

    Used by grant_types.
    Used by response_types:
        - ResponseTypeToken
    """

    expires_in: int
    refresh_token_expires_in: int
    access_token: str
    refresh_token: str
    scope: str
    token_type: str = "Bearer"


class TokenActiveIntrospectionResponse(NamedTuple):
    """Response for a valid access token.

    Used by token introspection endpoint.
    """

    scope: str
    client_id: str
    exp: int
    active: bool = True


class TokenInactiveIntrospectionResponse(NamedTuple):
    """For an invalid, revoked or expired token.

    Used by token introspection endpoint.
    """

    active: bool = False


class Response(NamedTuple):
    """General response class.

    Used by:
        - Endpoint
    """

    content: Optional[
        Union[
            ErrorResponse,
            TokenResponse,
            AuthorizationCodeResponse,
            TokenActiveIntrospectionResponse,
            TokenInactiveIntrospectionResponse,
        ]
    ] = None
    status_code: HTTPStatus = HTTPStatus.OK
    headers: CaseInsensitiveDict = default_headers



from typing import Dict, Optional, Type, Union

from ..grant_type import GrantTypeBase
from ..response_type import ResponseTypeBase
from ..types import EndpointType, GrantType, ResponseType
from .database import BaseDB


class BaseEndpoint:
    response_type: Dict[Optional[ResponseType], Type[ResponseTypeBase]] = {}
    grant_type: Dict[Optional[GrantType], Type[GrantTypeBase]] = {}
    available: bool = True

    def __init__(
        self, db: BaseDB, available: Optional[bool] = None,
    ):
        self.db = db

        if available is not None:
            self.available = available

    def register(
        self,
        endpoint_type: EndpointType,
        endpoint: Union[ResponseType, GrantType],
        endpoint_cls: Union[Type[ResponseTypeBase], Type[GrantTypeBase]],
    ):
        endpoint_dict = getattr(self, endpoint_type)
        endpoint_dict[endpoint] = endpoint_cls

    def unregister(
        self, endpoint_type: EndpointType, endpoint: Union[ResponseType, GrantType]
    ):
        endpoint_dict = getattr(self, endpoint_type)
        del endpoint_dict[endpoint]



from enum import Enum


class ErrorType(str, Enum):
    INVALID_REQUEST = "invalid_request"
    INVALID_CLIENT = "invalid_client"
    INVALID_GRANT = "invalid_grant"
    INVALID_SCOPE = "invalid_scope"
    UNAUTHORIZED_CLIENT = "unauthorized_client"
    UNSUPPORTED_GRANT_TYPE = "unsupported_grant_type"
    UNSUPPORTED_RESPONSE_TYPE = "unsupported_response_type"
    INSECURE_TRANSPORT = "insecure_transport"
    MISMATCHING_STATE = "mismatching_state"
    METHOD_IS_NOT_ALLOWED = "method_is_not_allowed"
    SERVER_ERROR = "server_error"
    TEMPORARILY_UNAVAILABLE = "temporarily_unavailable"


class GrantType(str, Enum):
    TYPE_AUTHORIZATION_CODE = "authorization_code"
    TYPE_PASSWORD = "password"
    TYPE_CLIENT_CREDENTIALS = "client_credentials"
    TYPE_REFRESH_TOKEN = "refresh_token"


class ResponseType(str, Enum):
    TYPE_TOKEN = "token"
    TYPE_CODE = "code"


class EndpointType(str, Enum):
    GRANT_TYPE = "grant_type"
    RESPONSE_TYPE = "response_type"


class RequestMethod(str, Enum):
    GET = "GET"
    POST = "POST"


class CodeChallengeMethod(str, Enum):
    PLAIN = "plain"
    S256 = "S256"



from http import HTTPStatus
from typing import Optional
from urllib.parse import urljoin

from .config import get_settings
from .constances import default_headers
from .structures import CaseInsensitiveDict
from .types import ErrorType


class OAuth2Error(Exception):
    error: ErrorType
    description: str = ""
    status_code: HTTPStatus = HTTPStatus.BAD_REQUEST
    error_uri: str = ""
    headers: CaseInsensitiveDict = default_headers

    def __init__(
        self,
        description: Optional[str] = None,
        headers: Optional[CaseInsensitiveDict] = None,
    ):
        settings = get_settings()

        if description is not None:
            self.description = description

        if headers is not None:
            self.headers = headers

        if settings.ERROR_URI:
            self.error_uri = urljoin(settings.ERROR_URI, self.error)

        super().__init__(f"({self.error}) {self.description}")


class MethodNotAllowedError(OAuth2Error):
    description = "HTTP method is not allowed."
    status_code: HTTPStatus = HTTPStatus.METHOD_NOT_ALLOWED
    error = ErrorType.METHOD_IS_NOT_ALLOWED


class InvalidRequestError(OAuth2Error):
    """
    The request is missing a required parameter, includes an invalid
    parameter value, includes a parameter more than once, or is
    otherwise malformed.
    """

    error = ErrorType.INVALID_REQUEST


class InvalidClientError(OAuth2Error):
    """
    Client authentication failed (e.g. unknown client, no client
    authentication included, or unsupported authentication method).
    The authorization server MAY return an HTTP 401 (Unauthorized) status
    code to indicate which HTTP authentication schemes are supported.
    If the client attempted to authenticate via the "Authorization" request
    header field, the authorization server MUST respond with an
    HTTP 401 (Unauthorized) status code, and include the "WWW-Authenticate"
    response header field matching the authentication scheme used by the
    client.
    """

    error = ErrorType.INVALID_CLIENT
    status_code: HTTPStatus = HTTPStatus.UNAUTHORIZED


class InsecureTransportError(OAuth2Error):
    description = "OAuth 2 MUST utilize https."
    error = ErrorType.INSECURE_TRANSPORT


class UnsupportedGrantTypeError(OAuth2Error):
    """
    The authorization grant type is not supported by the authorization
    server.
    """

    error = ErrorType.UNSUPPORTED_GRANT_TYPE


class UnsupportedResponseTypeError(OAuth2Error):
    """
    The authorization server does not support obtaining an authorization
    code using this method.
    """

    error = ErrorType.UNSUPPORTED_RESPONSE_TYPE


class InvalidGrantError(OAuth2Error):
    """
    The provided authorization grant (e.g. authorization code, resource
    owner credentials) or refresh token is invalid, expired, revoked, does
    not match the redirection URI used in the authorization request, or was
    issued to another client.

    https://tools.ietf.org/html/rfc6749#section-5.2
    """

    error = ErrorType.INVALID_GRANT


class MismatchingStateError(OAuth2Error):
    description = "CSRF Warning! State not equal in request and response."
    error = ErrorType.MISMATCHING_STATE


class UnauthorizedClientError(OAuth2Error):
    """
    The authenticated client is not authorized to use this authorization
    grant type.
    """

    error = ErrorType.UNAUTHORIZED_CLIENT


class InvalidScopeError(OAuth2Error):
    """
    The requested scope is invalid, unknown, or malformed, or
    exceeds the scope granted by the resource owner.

    https://tools.ietf.org/html/rfc6749#section-5.2
    """

    error = ErrorType.INVALID_SCOPE


class ServerError(OAuth2Error):
    """
    The authorization server encountered an unexpected condition that
    prevented it from fulfilling the request.  (This error code is needed
    because a 500 Internal Server Error HTTP status code cannot be returned
    to the client via a HTTP redirect.)
    """

    error = ErrorType.SERVER_ERROR


class TemporarilyUnavailableError(OAuth2Error):
    """
    The authorization server is currently unable to handle the request
    due to a temporary overloading or maintenance of the server.
    (This error code is needed because a 503 Service Unavailable HTTP
    status code cannot be returned to the client via a HTTP redirect.)
    """

    error = ErrorType.TEMPORARILY_UNAVAILABLE



from collections import UserDict


class CaseInsensitiveDict(UserDict):
    """A case-insensitive ``dict``-like object."""

    def __setitem__(self, key, value):
        super().__setitem__(key.lower(), value)

    def __getitem__(self, key):
        return super().__getitem__(key.lower())



from http import HTTPStatus

from .base.endpoint import BaseEndpoint
from .constances import default_headers
from .grant_type import GrantTypeBase
from .requests import Request
from .response_type import ResponseTypeBase
from .responses import (
    Response,
    TokenActiveIntrospectionResponse,
    TokenInactiveIntrospectionResponse,
)
from .structures import CaseInsensitiveDict
from .types import ResponseType
from .utils import build_uri, catch_errors_and_unavailability, decode_auth_headers


class Endpoint(BaseEndpoint):
    @catch_errors_and_unavailability
    async def create_token_introspection_response(self, request: Request) -> Response:
        client_id, _ = decode_auth_headers(request.headers.get("Authorization", ""))

        token = await self.db.get_token(
            request=request, client_id=client_id, token=request.post.token
        )

        token_response = TokenInactiveIntrospectionResponse()

        if token and not token.is_expired:
            token_response = TokenActiveIntrospectionResponse(
                scope=token.scope, client_id=token.client_id, exp=token.expires_in
            )

        return Response(
            content=token_response, status_code=HTTPStatus.OK, headers=default_headers
        )

    @catch_errors_and_unavailability
    async def create_token_response(self, request: Request) -> Response:
        grant_type_cls = self.grant_type.get(request.post.grant_type, GrantTypeBase)
        grant_type_handler = grant_type_cls(self.db)
        token_response = await grant_type_handler.create_token_response(request)
        return Response(
            content=token_response, status_code=HTTPStatus.OK, headers=default_headers
        )

    @catch_errors_and_unavailability
    async def create_authorization_code_response(self, request: Request) -> Response:
        response_type_cls = self.response_type.get(
            request.query.response_type, ResponseTypeBase
        )
        response_type_handler = response_type_cls(self.db)
        authorization_code_response = await response_type_handler.create_authorization_code_response(
            request
        )
        response_type = request.query.response_type
        response_dict = {
            **authorization_code_response._asdict(),
            "state": request.query.state,
        }
        query_params = response_dict if response_type == ResponseType.TYPE_CODE else {}
        fragment = response_dict if response_type == ResponseType.TYPE_TOKEN else {}

        location = build_uri(request.query.redirect_uri, query_params, fragment)

        return Response(
            status_code=HTTPStatus.FOUND,
            headers=CaseInsensitiveDict({"location": location}),
        )



import os
from typing import Any, Callable, NamedTuple


def get_env(env: str, default_value: Any, to_type: Callable, prefix="AIOAUTH_"):
    """Get the value of an environment variable and apply a specific type to it"""
    return to_type(os.environ.get(f"{prefix}{env}", default_value))


class Settings(NamedTuple):
    TOKEN_EXPIRES_IN: int
    AUTHORIZATION_CODE_EXPIRES_IN: int
    INSECURE_TRANSPORT: bool
    ERROR_URI: str


def get_settings():
    return Settings(
        TOKEN_EXPIRES_IN=get_env("TOKEN_EXPIRES_IN", 86400, int),
        AUTHORIZATION_CODE_EXPIRES_IN=get_env(
            "AUTHORIZATION_CODE_EXPIRES_IN", 300, int
        ),
        INSECURE_TRANSPORT=get_env("INSECURE_TRANSPORT", False, bool),
        ERROR_URI=get_env("ERROR_URI", "", str),
    )



from typing import Optional

from .base.request_validator import BaseRequestValidator
from .errors import (
    InvalidClientError,
    InvalidRequestError,
    InvalidScopeError,
    UnsupportedResponseTypeError,
)
from .models import Client
from .requests import Request
from .responses import AuthorizationCodeResponse, TokenResponse
from .types import CodeChallengeMethod, RequestMethod, ResponseType


class ResponseTypeBase(BaseRequestValidator):
    response_type: Optional[ResponseType] = None
    allowed_methods = [
        RequestMethod.GET,
    ]
    code_challenge_methods = list(CodeChallengeMethod)

    async def validate_request(self, request: Request) -> Client:
        await super().validate_request(request)

        if not request.query.client_id:
            raise InvalidRequestError(description="Missing client_id parameter.")

        client = await self.db.get_client(
            request=request, client_id=request.query.client_id
        )

        if not client:
            raise InvalidRequestError(description="Invalid client_id parameter value.")

        if not request.query.redirect_uri:
            raise InvalidRequestError(description="Mismatching redirect URI.")

        if self.response_type != request.query.response_type:
            raise UnsupportedResponseTypeError()

        if not client.check_redirect_uri(request.query.redirect_uri):
            raise InvalidRequestError(description="Invalid redirect URI.")

        if not request.query.response_type:
            raise InvalidRequestError(description="Missing response_type parameter.")

        if request.query.code_challenge_method:
            if request.query.code_challenge_method not in self.code_challenge_methods:
                raise InvalidRequestError(
                    description="Transform algorithm not supported."
                )

            if not request.query.code_challenge:
                raise InvalidRequestError(description="Code challenge required.")

        if not client.check_response_type(request.query.response_type):
            raise UnsupportedResponseTypeError()

        if not client.check_scope(request.query.scope):
            raise InvalidScopeError()

        if not request.user:
            raise InvalidClientError(description="User is not authorized")

        return client

    async def create_authorization_code_response(self, request: Request) -> Client:
        return await self.validate_request(request)


class ResponseTypeToken(ResponseTypeBase):
    response_type: ResponseType = ResponseType.TYPE_TOKEN

    async def create_authorization_code_response(
        self, request: Request
    ) -> TokenResponse:
        client = await super().create_authorization_code_response(request)
        token = await self.db.create_token(
            request, client.client_id, request.query.scope
        )
        return TokenResponse(
            expires_in=token.expires_in,
            refresh_token_expires_in=token.refresh_token_expires_in,
            access_token=token.access_token,
            refresh_token=token.refresh_token,
            scope=token.scope,
            token_type=token.token_type,
        )


class ResponseTypeAuthorizationCode(ResponseTypeBase):
    response_type: ResponseType = ResponseType.TYPE_CODE

    async def create_authorization_code_response(
        self, request: Request
    ) -> AuthorizationCodeResponse:
        client = await super().create_authorization_code_response(request)
        authorization_code = await self.db.create_authorization_code(
            request,
            client.client_id,
            request.query.scope,
            request.query.response_type,  # type: ignore
            request.query.redirect_uri,
            request.query.code_challenge_method,  # type: ignore
            request.query.code_challenge,  # type: ignore
        )
        return AuthorizationCodeResponse(
            code=authorization_code.code, scope=authorization_code.scope,
        )



from typing import Optional

from .base.request_validator import BaseRequestValidator
from .errors import (
    InvalidGrantError,
    InvalidRequestError,
    InvalidScopeError,
    MismatchingStateError,
    UnauthorizedClientError,
    UnsupportedGrantTypeError,
)
from .models import Client
from .requests import Request
from .responses import TokenResponse
from .types import GrantType, RequestMethod
from .utils import decode_auth_headers


class GrantTypeBase(BaseRequestValidator):
    allowed_methods = [
        RequestMethod.POST,
    ]
    grant_type: Optional[GrantType] = None

    async def create_token_response(self, request: Request) -> TokenResponse:
        client = await self.validate_request(request)
        token = await self.db.create_token(
            request, client.client_id, request.post.scope
        )

        return TokenResponse(
            expires_in=token.expires_in,
            refresh_token_expires_in=token.refresh_token_expires_in,
            access_token=token.access_token,
            refresh_token=token.refresh_token,
            scope=token.scope,
            token_type=token.token_type,
        )

    async def validate_request(self, request: Request) -> Client:
        await super().validate_request(request)

        client_id, client_secret = decode_auth_headers(
            request.headers.get("Authorization", "")
        )

        if not request.post.grant_type:
            raise InvalidRequestError(description="Request is missing grant type.")

        if self.grant_type != request.post.grant_type:
            raise UnsupportedGrantTypeError()

        client = await self.db.get_client(
            request, client_id=client_id, client_secret=client_secret
        )

        if not client:
            raise InvalidRequestError(description="Invalid client_id parameter value.")

        if not client.check_grant_type(request.post.grant_type):
            raise UnauthorizedClientError()

        if not client.check_scope(request.post.scope):
            raise InvalidScopeError()

        return client


class AuthorizationCodeGrantType(GrantTypeBase):
    grant_type: GrantType = GrantType.TYPE_AUTHORIZATION_CODE

    async def validate_request(self, request: Request) -> Client:
        client = await super().validate_request(request)

        if not request.post.redirect_uri:
            raise InvalidRequestError(description="Mismatching redirect URI.")

        if not client.check_redirect_uri(request.post.redirect_uri):
            raise InvalidRequestError(description="Invalid redirect URI.")

        if not request.post.code:
            raise InvalidRequestError(description="Missing code parameter.")

        authorization_code = await self.db.get_authorization_code(
            request, client.client_id, request.post.code
        )

        if not authorization_code:
            raise InvalidGrantError()

        if (
            authorization_code.code_challenge
            and authorization_code.code_challenge_method
        ):
            if not request.post.code_verifier:
                raise InvalidRequestError(description="Code verifier required.")

            is_valid_code_challenge = authorization_code.check_code_challenge(
                request.post.code_verifier
            )
            if not is_valid_code_challenge:
                raise MismatchingStateError()

        if authorization_code.is_expired:
            raise InvalidGrantError()

        await self.db.delete_authorization_code(
            request, client.client_id, request.post.code
        )

        return client


class PasswordGrantType(GrantTypeBase):
    grant_type: GrantType = GrantType.TYPE_PASSWORD

    async def validate_request(self, request: Request) -> Client:
        client = await super().validate_request(request)

        if not request.post.password or not request.post.password:
            raise InvalidGrantError(description="Invalid credentials given.")

        user = await self.db.authenticate(request)

        if not user:
            raise InvalidGrantError(description="Invalid credentials given.")

        return client


class RefreshTokenGrantType(GrantTypeBase):
    grant_type: GrantType = GrantType.TYPE_REFRESH_TOKEN

    async def validate_request(self, request: Request) -> Client:
        client = await super().validate_request(request)

        if not request.post.refresh_token:
            raise InvalidRequestError(description="Missing refresh token parameter.")

        token = await self.db.get_token(
            request=request,
            client_id=client.client_id,
            refresh_token=request.post.refresh_token,
        )

        if not token:
            raise InvalidGrantError()

        if token.refresh_token_expired:
            raise InvalidGrantError()

        await self.db.revoke_token(request, request.post.refresh_token)

        return client


class ClientCredentialsGrantType(GrantTypeBase):
    grant_type: GrantType = GrantType.TYPE_CLIENT_CREDENTIALS



from aioauth.structures import CaseInsensitiveDict

from ..constances import default_headers
from ..errors import InsecureTransportError, MethodNotAllowedError
from ..requests import Request
from ..types import RequestMethod
from ..utils import is_secure_transport
from .database import BaseDB


class BaseRequestValidator:
    allowed_methods = [
        RequestMethod.GET,
        RequestMethod.POST,
    ]

    def __init__(self, db: BaseDB):
        self.db = db

    async def validate_request(self, request: Request):
        if not is_secure_transport(request.url):
            raise InsecureTransportError()

        if request.method not in self.allowed_methods:
            headers = CaseInsensitiveDict(
                {**default_headers, "allow": ", ".join(self.allowed_methods)}
            )
            raise MethodNotAllowedError(headers=headers)



from .structures import CaseInsensitiveDict


def _default_headers() -> CaseInsensitiveDict:
    """The authorization server MUST include the HTTP "Cache-Control"
    response header field [RFC2616] with a value of "no-store" in any
    response containing tokens, credentials, or other sensitive
    information, as well as the "Pragma" response header field [RFC2616]
    with a value of "no-cache".
    """
    return CaseInsensitiveDict(
        {
            "Content-Type": "application/json",
            "Cache-Control": "no-store",
            "Pragma": "no-cache",
        }
    )


default_headers = _default_headers()



import time
from typing import Optional

from aioauth.types import CodeChallengeMethod, ResponseType

from ..config import get_settings
from ..models import AuthorizationCode, Client, Token
from ..requests import Request
from ..utils import generate_token


class BaseDB:
    async def create_token(self, request: Request, client_id: str, scope: str) -> Token:
        """Generates Token model instance.

        Generated Token MUST be stored in database.

        Method is used by all core grant types.
        Method is used by response types:
            - ResponseTypeToken
        """
        settings = get_settings()
        return Token(
            client_id=client_id,
            expires_in=settings.TOKEN_EXPIRES_IN,
            access_token=generate_token(42),
            refresh_token=generate_token(48),
            issued_at=int(time.time()),
            scope=scope,
            revoked=False,
        )

    async def get_token(
        self,
        request: Request,
        client_id: str,
        token: Optional[str] = None,
        refresh_token: Optional[str] = None,
    ) -> Optional[Token]:
        """Gets existing token from the database

        Method is used by:
            - create_token_introspection_response
        Method is used by grant types:
            - RefreshTokenGrantType
        """
        raise NotImplementedError("Method get_token must be implemented")

    async def create_authorization_code(
        self,
        request: Request,
        client_id: str,
        scope: str,
        response_type: ResponseType,
        redirect_uri: str,
        code_challenge_method: CodeChallengeMethod,
        code_challenge: str,
    ) -> AuthorizationCode:
        """Generates AuthorizationCode model instance.

        Generated AuthorizationCode MUST be stored in database.

        Method is used by response types:
            - ResponseTypeAuthorizationCode
        """
        return AuthorizationCode(
            code=generate_token(48),
            client_id=client_id,
            redirect_uri=redirect_uri,
            response_type=response_type,
            scope=scope,
            auth_time=int(time.time()),
            code_challenge_method=code_challenge_method,
            code_challenge=code_challenge,
        )

    async def get_client(
        self, request: Request, client_id: str, client_secret: Optional[str] = None
    ) -> Optional[Client]:
        """Gets existing Client from database.

        If client doesn't exists in database this method MUST return None
        to indicate to the validator that the requested ``client_id`` does not exist or is invalid.

        Method is used by all core grant types.
        Method is used by all core response types.
        """
        raise NotImplementedError("Method get_client must be implemented")

    async def authenticate(self, request: Request) -> bool:
        """Authenticate user.

        Method is used by grant types:
            - PasswordGrantType
        """
        raise NotImplementedError("Method authenticate must be implemented")

    async def get_authorization_code(
        self, request: Request, client_id: str, code: str
    ) -> Optional[AuthorizationCode]:
        """Gets existing AuthorizationCode from database.

        If authorization code doesn't exists it MUST return None
        to indicate to the validator that the requested authorization code does not exist or is invalid.

        Method is used by grant types:
            - AuthorizationCodeGrantType
        """
        raise NotImplementedError(
            "Method get_authorization_code must be implemented for AuthorizationCodeGrantType"
        )

    async def delete_authorization_code(
        self, request: Request, client_id: str, code: str
    ):
        """Deletes authorization code from database.

        Method is used by grant types:
            - AuthorizationCodeGrantType
        """
        raise NotImplementedError(
            "Method delete_authorization_code must be implemented for AuthorizationCodeGrantType"
        )

    async def revoke_token(self, request: Request, token: str) -> None:
        """Revokes token in database.

        This method MUST set `revoked` in True for existing token record.

        Method is used by grant types:
            - RefreshTokenGrantType
        """
        raise NotImplementedError(
            "Method revoke_token must be implemented for RefreshTokenGrantType"
        )



from pathlib import Path

from setuptools import find_packages, setup

here = Path(__file__).parent
about = {}

with open(here / "src" / "aioauth" / "__version__.py", "r") as f:
    exec(f.read(), about)

with open("README.md") as readme_file:
    readme = readme_file.read()


def read_requirements(path):
    try:
        with path.open(mode="rt", encoding="utf-8") as fp:
            return list(filter(bool, (line.split("#")[0].strip() for line in fp)))
    except IndexError:
        raise RuntimeError(f"{path} is broken")


base_requirements = read_requirements(here / "requirements" / "base.txt")
test_requirements = read_requirements(here / "requirements" / "test.txt")

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    license=about["__license__"],
    python_requires=">=3.6.0",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=base_requirements,
    tests_require=test_requirements,
    extras_require={"test": test_requirements},
    include_package_data=True,
    keywords="asyncio oauth2 oauth",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    test_suite="tests",
    zip_safe=False,
    project_urls={"Source": about["__url__"]},
)



from typing import Any, NamedTuple, Optional

from .structures import CaseInsensitiveDict
from .types import CodeChallengeMethod, GrantType, RequestMethod, ResponseType


class Query(NamedTuple):
    client_id: Optional[str] = None
    redirect_uri: str = ""
    response_type: Optional[ResponseType] = None
    state: str = ""
    scope: str = ""
    code_challenge_method: Optional[CodeChallengeMethod] = None
    code_challenge: Optional[str] = None


class Post(NamedTuple):
    grant_type: Optional[GrantType] = None
    redirect_uri: Optional[str] = None
    scope: str = ""
    username: Optional[str] = None
    password: Optional[str] = None
    refresh_token: Optional[str] = None
    code: Optional[str] = None
    token: Optional[str] = None
    code_verifier: Optional[str] = None


class Request(NamedTuple):
    method: RequestMethod
    headers: CaseInsensitiveDict = CaseInsensitiveDict()
    query: Query = Query()
    post: Post = Post()
    url: str = ""
    user: Optional[Any] = None



import time
from typing import List, NamedTuple, Optional, Text

from .config import get_settings
from .types import CodeChallengeMethod, GrantType, ResponseType
from .utils import create_s256_code_challenge, list_to_scope, scope_to_list


class ClientMetadata(NamedTuple):
    grant_types: List[GrantType] = []
    response_types: List[ResponseType] = []
    redirect_uris: List[str] = []
    scope: Text = ""


class Client(NamedTuple):
    client_id: Text
    client_secret: Text
    client_metadata: ClientMetadata

    def check_redirect_uri(self, redirect_uri) -> bool:
        return redirect_uri in self.client_metadata.redirect_uris

    def check_grant_type(self, grant_type: GrantType) -> bool:
        return grant_type in self.client_metadata.grant_types

    def check_response_type(self, response_type: ResponseType) -> bool:
        return response_type in self.client_metadata.response_types

    def get_allowed_scope(self, scope) -> Text:
        if not scope:
            return ""
        allowed = set(self.client_metadata.scope.split())
        scopes = scope_to_list(scope)
        return list_to_scope([s for s in scopes if s in allowed])

    def check_scope(self, scope: str) -> bool:
        allowed_scope = self.get_allowed_scope(scope)
        return not (set(scope_to_list(scope)) - set(scope_to_list(allowed_scope)))


class AuthorizationCode(NamedTuple):
    code: Text
    client_id: Text
    redirect_uri: Text
    response_type: ResponseType
    scope: Text
    auth_time: int
    code_challenge: Optional[Text] = None
    code_challenge_method: Optional[CodeChallengeMethod] = None
    nonce: Optional[Text] = None

    def check_code_challenge(self, code_verifier: str) -> bool:
        is_valid_code_challenge = False

        if self.code_challenge_method == CodeChallengeMethod.PLAIN:
            # If the "code_challenge_method" was "plain", they are compared directly
            is_valid_code_challenge = code_verifier == self.code_challenge

        if self.code_challenge_method == CodeChallengeMethod.S256:
            # base64url(sha256(ascii(code_verifier))) == code_challenge
            is_valid_code_challenge = (
                create_s256_code_challenge(code_verifier) == self.code_challenge
            )

        return is_valid_code_challenge

    @property
    def is_expired(self) -> bool:
        settings = get_settings()
        return self.auth_time + settings.AUTHORIZATION_CODE_EXPIRES_IN < time.time()


class Token(NamedTuple):
    access_token: Text
    refresh_token: Text
    scope: Text
    issued_at: int
    expires_in: int
    client_id: Text
    token_type: Text = "Bearer"
    revoked: bool = False

    @property
    def is_expired(self) -> bool:
        return self.token_expires_in < time.time()

    @property
    def refresh_token_expires_in(self) -> int:
        expires_at = self.issued_at + self.expires_in * 2
        return expires_at

    @property
    def token_expires_in(self) -> int:
        expires_at = self.issued_at + self.expires_in
        return expires_at

    @property
    def refresh_token_expired(self) -> bool:
        return self.refresh_token_expires_in < time.time()



__title__ = "aioauth"
__description__ = "Asynchronous OAuth 2.0 framework for Python 3."
__url__ = "https://github.com/aliev/aioauth"
__version__ = "0.1.1"
__author__ = "Ali Aliyev"
__author_email__ = "ali@aliev.me"
__license__ = "The MIT License (MIT)"
__copyright__ = "Copyright 2020 Ali Aliyev"



import base64
import binascii
import functools
import hashlib
import logging
import random
import string
from base64 import b64decode, b64encode
from typing import Callable, Dict, List, Optional, Set, Text, Tuple, Union
from urllib.parse import quote, urlencode, urlparse, urlunsplit

from .config import get_settings
from .errors import (
    InvalidClientError,
    OAuth2Error,
    ServerError,
    TemporarilyUnavailableError,
)
from .responses import ErrorResponse, Response
from .structures import CaseInsensitiveDict

UNICODE_ASCII_CHARACTER_SET = string.ascii_letters + string.digits


log = logging.getLogger(__name__)


def is_secure_transport(uri: str) -> bool:
    """Check if the uri is over ssl."""
    settings = get_settings()

    if settings.INSECURE_TRANSPORT:
        return True
    return uri.lower().startswith("https://")


def get_authorization_scheme_param(
    authorization_header_value: Text,
) -> Tuple[Text, Text]:
    if not authorization_header_value:
        return "", ""
    scheme, _, param = authorization_header_value.partition(" ")
    return scheme, param


def list_to_scope(scope: Optional[List] = None) -> Text:
    """Convert a list of scopes to a space separated string."""
    if isinstance(scope, str) or scope is None:
        return ""
    elif isinstance(scope, (set, tuple, list)):
        return " ".join([str(s) for s in scope])
    else:
        raise ValueError(
            "Invalid scope (%s), must be string, tuple, set, or list." % scope
        )


def scope_to_list(scope: Union[Text, List, Set, Tuple]) -> List:
    """Convert a space separated string to a list of scopes."""
    if isinstance(scope, (tuple, list, set)):
        return [str(s) for s in scope]
    elif scope is None:
        return []
    else:
        return scope.strip().split(" ")


def generate_token(length: int = 30, chars: str = UNICODE_ASCII_CHARACTER_SET) -> str:
    """Generates a non-guessable OAuth token

    OAuth (1 and 2) does not specify the format of tokens except that they
    should be strings of random characters. Tokens should not be guessable
    and entropy when generating the random characters is important. Which is
    why SystemRandom is used instead of the default random.choice method.
    """
    rand = random.SystemRandom()
    return "".join(rand.choice(chars) for _ in range(length))


def build_uri(
    url: str, query_params: Optional[Dict] = None, fragment: Optional[Dict] = None
) -> str:
    """Build uri string from given url, query_params and fragment"""
    if query_params is None:
        query_params = {}

    if fragment is None:
        fragment = {}

    parsed_url = urlparse(url)
    uri = urlunsplit(
        (
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            urlencode(query_params, quote_via=quote),
            urlencode(fragment, quote_via=quote),
        )
    )
    return uri


def encode_auth_headers(client_id: str, client_secret: str) -> CaseInsensitiveDict:
    authorization = b64encode(f"{client_id}:{client_secret}".encode("ascii"))
    return CaseInsensitiveDict(Authorization=f"basic {authorization.decode()}")


def decode_auth_headers(authorization: str) -> Tuple[str, str]:
    """Decode an encrypted HTTP basic authentication string. Returns a tuple of
    the form (client_id, client_secret), and raises a InvalidClientError exception if
    nothing could be decoded.
    """
    headers = CaseInsensitiveDict({"WWW-Authenticate": "Basic"})

    scheme, param = get_authorization_scheme_param(authorization)
    if not authorization or scheme.lower() != "basic":
        raise InvalidClientError(headers=headers)

    try:
        data = b64decode(param).decode("ascii")
    except (ValueError, UnicodeDecodeError, binascii.Error):
        raise InvalidClientError(headers=headers)

    client_id, separator, client_secret = data.partition(":")

    if not separator:
        raise InvalidClientError(headers=headers)

    return client_id, client_secret


def create_s256_code_challenge(code_verifier: str) -> str:
    """Create S256 code_challenge with the given code_verifier.

    Implements:
        base64url(sha256(ascii(code_verifier)))
    """
    code_verifier_bytes = code_verifier.encode("utf-8")
    data = hashlib.sha256(code_verifier_bytes).digest()
    return base64.urlsafe_b64encode(data).rstrip(b"=").decode()


def catch_errors_and_unavailability(f) -> Callable:
    @functools.wraps(f)
    async def wrapper(endpoint, *args, **kwargs) -> Optional[Response]:
        if not endpoint.available:
            error = TemporarilyUnavailableError()
            content = ErrorResponse(error=error.error, description=error.description)
            return Response(
                content=content, status_code=error.status_code, headers=error.headers
            )

        try:
            response = await f(endpoint, *args, **kwargs)
            return response
        except OAuth2Error as exc:
            content = ErrorResponse(error=exc.error, description=exc.description)
            log.debug(exc)
            return Response(
                content=content, status_code=exc.status_code, headers=exc.headers
            )
        except Exception:
            error = ServerError()
            log.exception("Exception caught while processing request.")
            content = ErrorResponse(error=error.error, description=error.description)
            return Response(
                content=content, status_code=error.status_code, headers=error.headers,
            )

    return wrapper
